/******************************************************************************
*
*  mp3tag.cmd
*
*  Synchronize file names and ID3 tags of MPEG audio files.
*  Or sort then into directories.
*
*  Create:  10.10.1999
*  Update:  21.08.2002
*  (C)      Marcel Mller 1999-2002
*
*  History:
*  07.07.2002:  Ogg Vorbis support (/fo)
*
******************************************************************************/

IF RxFuncAdd("SysLoadFuncs", "RexxUtil", "SysLoadFuncs") = 0 THEN
   CALL SysLoadFuncs

IF RxFuncAdd("MMLoadFuncs", "RxMMutl", "MMLoadFuncs") = 0 THEN
   CALL MMLoadFuncs

/* initialize */
opt.tsp = '00'x||SUBSTR(InitTSP(), 2) /* preserve the 0bytes, they are used as path seperators */

/* some defaults */
opt.scan = ''
opt.template = ''
opt.sync = 0
opt.ea = ''
opt.pm = 0
opt.caseadj = 0
opt.cp = 1
opt.keep30 = 1
opt.undersc = 1
opt.trunc = 0
opt.autotrack = 0
opt.ogg = 1
nspecs = 0

/* parse command line */
params = STRIP(ARG(1))
DO WHILE params \= ''
   IF LEFT(params,1) = '"' THEN DO
      PARSE VAR params '"'param'"'params
      END
    ELSE
      PARSE VAR params param params
   params = STRIP(params, 'L')
   IF VERIFY(LEFT(param, 1), '/-') \= 0 THEN DO
      /* filespec */
      nspecs = nspecs + 1
      spec.nspecs = param
      ITERATE
      END
   optn = SUBSTR(param, 2)
   SELECT
    WHEN optn = 's' THEN
      opt.scan = 'S'
    WHEN optn = 'n0' THEN
      opt.template = '%.%a\%t%~'
    WHEN optn = 'n1' THEN
      opt.template = '%.%a\%A\%T - %t%~'
    WHEN optn = 'n2' THEN
      opt.template = '%.%A\%T - %a - %t%~'
    WHEN ABBREV(optn, 'n') THEN
      opt.template = SUBSTR(param, 3)
    WHEN ABBREV(optn, 'ot') THEN
      opt.title = SUBSTR(param, 4)
    WHEN ABBREV(optn, 'oa') THEN
      opt.artist = SUBSTR(param, 4)
    WHEN ABBREV(optn, 'oA') THEN
      opt.album = SUBSTR(param, 4)
    WHEN ABBREV(optn, 'oT') THEN
      opt.track = SUBSTR(param, 4)
    WHEN ABBREV(optn, 'oy') THEN
      opt.year = SUBSTR(param, 4)
    WHEN ABBREV(optn, 'oc') THEN
      opt.comment = SUBSTR(param, 4)
    WHEN ABBREV(optn, 'os') THEN
      opt.style = SUBSTR(param, 4)
    WHEN optn = '!k' | optn = '\k' THEN
      opt.keep30 = 0
    WHEN optn = '!u' | optn = '\u' THEN
      opt.sync = -1
    WHEN optn = 'u' THEN
      opt.sync = 1
    WHEN optn = 'U' THEN
      opt.sync = 2
    WHEN optn = '_' THEN
      opt.undersc = 1
    WHEN optn = '!_' | optn = '\_' THEN
      opt.undersc = 0
    WHEN optn = 'ac' THEN
      opt.caseadj = 1
    WHEN optn = 'aT' THEN
      opt.autotrack = 1
    WHEN optn = '!cp' | optn = '\cp' THEN
      opt.cp = 0
    WHEN optn = 't' THEN
      opt.trunc = 1
    WHEN optn = '!fo' | optn = '\fo' THEN
      opt.ogg = 0
    WHEN ABBREV(optn, 'ea') THEN DO
      IF LENGTH(optn) = 2 THEN
         opt.ea = '*'
       ELSE
         opt.ea = SUBSTR(param, 4)
      END
    WHEN optn = '?' THEN DO
      SAY 'mp3TAG <filespec> [<options>]'
      SAY '(C) Marcel Mller 1999-2002'
      SAY
      SAY 'default filespec: *.mp? *.ogg'
      SAY 'options:'
      SAY ' /s           include Subdirectories'
      SAY ' /n<template> naming convention'
      SAY '   %a or *a   artist'
      SAY '   %t or *t   title'
      SAY '   %A or *A   album'
      SAY '   %T or *T   track (two digits)'
      SAY '   %y or *y   year'
      SAY '   %c or *c   comment'
      SAY '   %. or *.   current directory'
      SAY '   %\ or *\   directory of the file'
      SAY '   %$ or *$   filename without extension'
      SAY '   %~ or *~   file extension'
      SAY '   %* or **   dummy parameter (wildcard)'
      SAY ' /n0          = /n%.%a\%t%~ (default)'
      SAY ' /n1          = "/n%.%a\%A\%T - %t%~" (Album)'
      SAY ' /n2          = "/n%.%A\%T - %a - %t%~" (Sampler)'
      SAY ' /oa<artist>  overwrite artist'
      SAY ' /oA<album>   overwrite album'
      SAY ' /oy<year>    overwrite year'
      SAY ' /oc<comment> overwrite comment'
      SAY ' /os<style>   overwrite style (sorry, only by number)'
      SAY ' /u           update ID3 from filename, priority ID3'
      SAY ' /U           update ID3 from filename, priority filename'
      SAY ' /!_ or /\_   disable underscores to space conversion'
      SAY ' /ac          adjust case in names'
      SAY ' /!cp or /\cp no codepage translation'
      SAY ' /ea[<type>]  set .TYPE EA (default type: MP3)'
      SAY ' /pm          use PMShell for file move'
      SAY ' /!fo /\fo    disable ogg vorbis extension (featuring vorbiscomment)'
      EXIT 48
      END
    OTHERWISE
      CALL Error 44, 'bad option 'param
      END
  END

/* prepare */
IF nspecs = 0 THEN
   IF opt.sync = 0 THEN DO
      nspecs = 1 + opt.ogg
      spec.1 = '*.mp?'
      IF opt.ogg THEN
         spec.2 = '*.ogg'
      END
    ELSE
      CALL Error 40, 'Nothing to do! - Filename expected'
IF opt.sync > 0 THEN DO
   opt.nopath = 0
   opt.noext = 0
   opt.mask = '"'opt.template'"'
   /* escape "'" */
   /* parse template */
   p = 1
   DO FOREVER
      p = VERIFY(opt.mask, '%*', 'M', p)
      IF p = 0 THEN LEAVE
      c = SUBSTR(opt.mask, p+1, 1)
      SELECT
       WHEN c = 'a' THEN
         new = '"tag2.artist"'
       WHEN c = 't' THEN
         new = '"tag2.title"'
       WHEN c = 'A' THEN
         new = '"tag2.album"'
       WHEN c = 'T' THEN
         new = '"tag2.track"'
       WHEN c = 'y' THEN
         new = '"tag2.year"'
       WHEN c = 'c' THEN
         new = '"tag2.comment"'
       WHEN c = '.' THEN
         new = STRIP(DIRECTORY(),'T','\')'\'
       WHEN c = '\' THEN DO
         new = ''
         opt.nopath = 1
         END
       WHEN c = '$' THEN DO
         CALL Error 34, '%$ makes no sense with /u.'
         END
       WHEN c = '~' THEN DO
         new = ''
         opt.noext = 1
         END
       WHEN c = '%' THEN
         new = '%'
       WHEN c = '*' THEN
         new = '"dummy"'
       OTHERWISE
         p = p + 2
         ITERATE
         END
      /* replace */
      opt.mask = INSERT(new, DELSTR(opt.mask, p, 2), p -1)
      p = p + LENGTH(new)
      END
   /* replace '""' by ' ' */
   p = 1
   DO FOREVER
      p = POS('""', opt.mask, p)
      IF p = 0 THEN LEAVE
      opt.mask = OVERLAY('  ', opt.mask, p)
      END
   SAY "using parse mask "opt.mask
   END
 ELSE; IF opt.template = '' THEN
   opt.template = '%.%a\%t%~'


sum = 0

DO j = 1 TO nspecs
   /* search files */
   CALL SysFileTree spec.j, files, "OF"opt.scan, '*--*-'
   sum = sum + files.0

   DO i = 1 TO files.0
      IF opt.sync > 0 THEN
         CALL SyncFile files.i
       ELSE IF opt.sync = 0 THEN
         CALL RenFile files.i
       ELSE
         CALL DoFile files.i
      END

   END

IF sum = 0 THEN DO
   tmp = ''
   DO j = 1 TO nspecs
      tmp = tmp' 'spec.j
      END
   SAY "no matching files found for"tmp
   END

EXIT

DoFile: PROCEDURE EXPOSE opt.
   /* read header, read tag */
   fform = ReadTag(ARG(1))
   IF fform = '' THEN
      RETURN

   /* .TYPE EA */
   IF opt.ea = '*' THEN
      CALL SysPutEA ARG(1), '.TYPE', 'DFFF00000100FDFF'x||REVERSE(D2C(LENGTH(fform),2))||fform
    ELSE IF opt.ea \= '' THEN
      CALL SysPutEA ARG(1), '.TYPE', 'DFFF00000100FDFF'x||REVERSE(D2C(LENGTH(opt.ea),2))||opt.ea
   /* purge .CLASSINFO EA */
   CALL SysPutEA ARG(1), '.CLASSINFO', ''
   RETURN

SyncFile: PROCEDURE EXPOSE opt.
   /* read header, read tag */
   fform = ReadTag(ARG(1))
   IF fform = '' THEN
      RETURN

   /* priority: .LONGNAME */
   name = ToLongname(ARG(1))

   /* fetch from filename */
   IF opt.nopath THEN
      name = FILESPEC('N', name)
   IF opt.noext THEN DO
      p = LASTPOS('.', name)
      IF p \= 0 THEN
         name = DELSTR(name, p)
      END
   IF opt.undersc THEN
      name = TRANSLATE(name,, '_')
   INTERPRET "PARSE VAR name "opt.mask
   IF SYMBOL("tag2.track") \= 'VAR' & opt.autotrack THEN DO
      tag2.track = GetAutoTrack(name)
      IF tag2.track = '' THEN DROP tag2.track
      END

   /* merge data */
   CALL MergeData 'title'
   CALL MergeData 'artist'
   CALL MergeData 'album'
   CALL MergeData 'track'
   CALL MergeData 'year'
   CALL MergeData 'comment'

   /* option overwrites */
   CALL ID3overwrite

   /* .TYPE EA */
   IF opt.ea = '*' THEN
      CALL SysPutEA ARG(1), '.TYPE', 'DFFF00000100FDFF'x||REVERSE(D2C(LENGTH(fform),2))||fform
    ELSE IF opt.ea \= '' THEN
      CALL SysPutEA ARG(1), '.TYPE', 'DFFF00000100FDFF'x||REVERSE(D2C(LENGTH(opt.ea),2))||opt.ea

   /* write tag */
   CALL WriteTag ARG(1), fform

   RETURN

MergeData: PROCEDURE EXPOSE opt. tag. tag2.
   SELECT
    WHEN opt.sync = 1 THEN DO /* priority ID3 */
      IF SYMBOL('tag2.'ARG(1)) \= 'VAR' THEN RETURN
      val = STRIP(VALUE('tag2.'ARG(1)))
      IF SYMBOL('tag.'ARG(1)) \= 'VAR' THEN
         CALL SetTag ARG(1), Cp2ISO(val)
      END
    WHEN opt.sync = 2 THEN DO /* priority name */
      IF SYMBOL('tag2.'ARG(1)) \= 'VAR' THEN RETURN
      val = STRIP(VALUE('tag2.'ARG(1)))
      CALL SetTag ARG(1), Cp2ISO(val)
      END
    OTHERWISE NOP
      END
   RETURN

RenFile: PROCEDURE EXPOSE opt.
   /* read header, read tag */
   fform = ReadTag(ARG(1))
   IF fform = '' THEN
      RETURN

   /* try to fetch track otherwise */
   IF SYMBOL("tag.track") \= 'VAR' & opt.autotrack THEN DO
      tmp = FILESPEC('n', ARG(1))
      p = POS('.', tmp)
      IF p \= 0 THEN
         tmp = LEFT(tmp, p-1)
      tag.track = GetAutoTrack(tmp)
      IF tag.track = 0 THEN DO
         DROP tag.track
         p = POS('TRACK', TRANSLATE(tag.comment))
         IF p \= 0 THEN DO
            PARSE VALUE SUBSTR(tag.comment, p +5) WITH tmp dummy
            CALL SetTrack tmp
            END
         END
      END

   /* keep part of filename if ID3 tag seems to be truncated */
   IF opt.keep30 & fform = 'MP3' THEN DO
      name = ARG(1)
      IF opt.undersc THEN /* translate '_' */
         name = TRANSLATE(name,, '_')
      /* remove extension */
      p = LASTPOS('.', RIGHT(FILESPEC('N', name), 5))
      IF p > 0 THEN
         p = 6 - p
      name = LEFT(name, LENGTH(name) - p)
      /*SAY 'K: 'name*/
      /* check artist */
      IF LENGTH(tag.atrist) >= 29 THEN DO
         p = POS(TSP2(LEFT(tag.artist,30)), TSP2(name))
         IF p \= 0 THEN DO
            tmp = SUBSTR(name, p +30)
            q = MIN(POS('\', tmp'\'), POS(' - ', tmp' - '), POS(' / ', tmp' / '), POS('  ', tmp'  '))
            IF SUBSTR(name, p-1, 1) = '(' THEN
               q = MIN(q, POS(')', tmp')'))
            tag.artist = LEFT(tag.artist,30)||Cp2ISO(SUBSTR(name, p+30, q-1))
            END
         END
      /* check album */
      IF LENGTH(tag.album) >= 29 THEN DO
         p = POS(TSP2(LEFT(tag.album,30)), TSP2(name))
         IF p \= 0 THEN DO
            tmp = SUBSTR(name, p +30)
            q = MIN(POS('\', tmp'\'), POS(' - ', tmp' - '), POS(' / ', tmp' / '), POS('  ', tmp'  '))
            IF SUBSTR(name, p-1, 1) = '(' THEN
               q = MIN(q, POS(')', tmp')'))
            tag.album = LEFT(tag.album,30)||Cp2ISO(SUBSTR(name, p+30, q-1))
            END
         END
      /* check title */
      IF LENGTH(tag.title) >= 29 THEN DO
         name = FILESPEC('N', name)
         p = POS(TSP2(LEFT(tag.title,30)), TSP2(name))
         IF p \= 0 THEN DO
            tmp = SUBSTR(name, p +30)
            q = MIN(POS('\', tmp'\'), POS(' - ', tmp' - '), POS(' / ', tmp' / '), POS('  ', tmp'  '))
            tag.title = LEFT(tag.title,30)||Cp2ISO(SUBSTR(name, p+30, q-1))
            END
         END
      END

   /* option overwrites */
   CALL ID3overwrite

   /* .TYPE EA */
   IF opt.ea = '*' THEN
      CALL SysPutEA ARG(1), '.TYPE', 'DFFF00000100FDFF'x||REVERSE(D2C(LENGTH(fform),2))||fform
    ELSE IF opt.ea \= '' THEN
      CALL SysPutEA ARG(1), '.TYPE', 'DFFF00000100FDFF'x||REVERSE(D2C(LENGTH(opt.ea),2))||opt.ea

   /* parse template */
   name = TRANSLATE(opt.template, '0000'x, '\/')
   p = 1
   checked = 1
   DO FOREVER
      p = VERIFY(name, '%*', 'M', p)
      IF p = 0 THEN LEAVE
      c = SUBSTR(name, p+1, 1)
      SELECT
       WHEN c = 'a' THEN
         new = Cp2DOS(GetTag('artist'))
       WHEN c = 't' THEN
         new = Cp2DOS(GetTag('title'))
       WHEN c = 'A' THEN
         new = Cp2DOS(GetTag('album'))
       WHEN c = 'T' THEN
         new = Cp2DOS(GetTag('track'))
       WHEN c = 'y' THEN
         new = Cp2DOS(GetTag('year'))
       WHEN c = 'c' THEN
         new = Cp2DOS(GetTag('comment'))
       WHEN c = '.' THEN DO
         new = STRIP(DIRECTORY(),'T','\')'\'
         checked = p + LENGTH(new)
         END
       WHEN c = '00'x THEN DO /* translated from '\' */
         new = Path(ARG(1))
         checked = p + LENGTH(new)
         END
       WHEN c = '$' THEN
         new = Name(ARG(1))
       WHEN c = '~' THEN
         new = Extension(ARG(1))
       WHEN c = '%' THEN
         new = '%'
       WHEN c = '*' THEN
         new = ''
       OTHERWISE
         p = p + 2
         ITERATE
         END
      /* error ? */
      IF new = '' THEN SELECT
       WHEN c = 'a' THEN DO
         SAY "file "ARG(1)" has no artist information."
         RETURN
         END
       WHEN c = 't' THEN DO
         SAY "file "ARG(1)" has no title information."
         RETURN
         END
       WHEN c = 'A' THEN DO
         SAY "file "ARG(1)" has no album information."
         RETURN
         END
       WHEN c = 'T' THEN DO
         SAY "file "ARG(1)" has no track information."
         RETURN
         END
       OTHERWISE NOP
         END
      /* replace */
      name = INSERT(new, DELSTR(name, p, 2), p -1)
      p = p + LENGTH(new)
      END
   /* rename */
   CALL MoveTo ARG(1), name, checked
   RETURN

MoveTo: PROCEDURE EXPOSE opt.
/* ARG(1)    sourcefile
   ARG(2)    destination, path delimiters: '00'x
   ARG(3)    checked part of path
*/
   /*SAY "MT: "ARG(1) ARG(2) ARG(3)*/
   p = LASTPOS('00'x, ARG(2))
   /* check directory */
   dir = CheckDir(LEFT(ARG(2), p), ARG(3))
   IF dir = ':' THEN
      RETURN 0
   /*SAY "DD: "dir*/
   name = SUBSTR(ARG(2), p+1)
   /* move file */
   IF opt.pm THEN DO
      IF \SysMoveObject(ARG(1), dir||name) THEN DO
         SAY "file "ARG(1)" cannot be moved to "dir||name"."
         RETURN 0
         END
      RETURN 1
      END
   tname = TSP(name)
   /*SAY '@MOVE 'Escape(ARG(1))' 'Escape(FILESPEC('P', dir)||tname)' >NUL'*/
   '@MOVE 'Escape(ARG(1))' 'Escape(FILESPEC('P', dir)||tname)' >NUL'
   IF rc \= 0 THEN DO
      SAY "file "ARG(1)" cannot be moved to "dir||tname"("rc")."
      RETURN 0
      END
   CALL NameEA dir, tname, name
   SAY "file "ARG(1)" moved and/or renamed."
   RETURN 1

CheckDir: PROCEDURE EXPOSE opt.
/* ARG(1)    filename, path delimiters: '00'x
   ARG(2)    checked part of path
*/
   /*SAY "CHKD: "ARG(1) ARG(2)*/
   IF LENGTH(ARG(1)) < ARG(2) THEN
      RETURN ARG(1)
   name = LEFT(ARG(1), LENGTH(ARG(1)) -1)
   /* existence check */
   dir = LEFT(name, ARG(2)-1)TRANSLATE(TSP(SUBSTR(name, ARG(2))), '\', '00'x)
   IF STREAM(dir, 'c', 'query size') \= '' THEN /* query exists does not work for dirs */
      RETURN dir'\'
   p = LASTPOS('00'x, name)
   /* check directory */
   p = MAX(p, ARG(2)-1)
   dir = CheckDir(LEFT(name, p), ARG(2))
   /*SAY "D: "dir"_"*/
   IF dir = ':' THEN
      RETURN ':'
   name = SUBSTR(name, p+1)
   tname = TSP(name)
   /* create directory */
   IF SysMkDir(dir||tname) \= 0 THEN DO
      SAY 'Failed to create directory 'dir||tname'.'
      RETURN ':'
      END
   CALL NameEA dir, tname, name
   CALL SysSetObjectData dir||tname, 'SHOWALLINTREEVIEW=YES'
   RETURN dir||tname'\'

GetAutoTrack: PROCEDURE
   IF VERIFY(REVERSE(RIGHT(ARG(1), 3)), '0123456789') = 3 THEN
      RETURN RIGHT(ARG(1), 2)
   IF VERIFY(LEFT(ARG(1), 3), '0123456789') = 3 THEN
      RETURN LEFT(ARG(1), 2)
   RETURN ''

ID3overwrite: PROCEDURE EXPOSE opt. tag.
   /* automatic case adjustment */
   IF opt.caseadj THEN DO
      IF SYMBOL('tag.title') = 'VAR' THEN tag.title = CaseAdj(tag.title)
      IF SYMBOL('tag.artist') = 'VAR' THEN tag.artist = CaseAdj(tag.artist)
      IF SYMBOL('tag.album') = 'VAR' THEN tag.album = CaseAdj(tag.album)
      END
   /* direct overwrites */
   IF SYMBOL('opt.title') = 'VAR' THEN
      tag.title = Cp2ISO(opt.title)
   IF SYMBOL('opt.artist') = 'VAR' THEN
      tag.artist = Cp2ISO(opt.artist)
   IF SYMBOL('opt.album') = 'VAR' THEN
      tag.album = Cp2ISO(opt.album)
   IF SYMBOL('opt.year') = 'VAR' THEN
      tag.year = Cp2ISO(opt.year)
   IF SYMBOL('opt.track') = 'VAR' THEN
      tag.track = opt.track
   IF SYMBOL('opt.comment') = 'VAR' THEN
      tag.comment = Cp2ISO(opt.comment)
   IF SYMBOL('opt.style') = 'VAR' THEN
      tag.style = opt.style
   RETURN

GetTag: PROCEDURE EXPOSE tag.
   /* ARG(1) = tag entry
      ARG(2) = default
   */
   IF SYMBOL('tag.'ARG(1)) \= 'VAR' THEN RETURN ARG(2)
   RETURN VALUE('tag.'ARG(1))

SetTag: PROCEDURE EXPOSE tag.
   /* ARG(1) = tag entry
      ARG(2) = value
   */
   p = POS('0'x, ARG(2)'0'x)
   val = STRIP(LEFT(ARG(2), p-1))
   IF val = '' THEN RETURN
   CALL VALUE 'tag.'ARG(1), val
   RETURN

SetTrack: PROCEDURE EXPOSE tag.
   /* ARG(1) = track no
      return 0 = no success
             1 = OK
   */
   IF \DATATYPE(ARG(1), 'W') THEN RETURN 0
   CALL SetTag 'track', RIGHT(ARG(1), 2, '0')
   RETURN 1

GetTagOption: PROCEDURE EXPOSE tag.
   /* ARG(1) tag name
      ARG(2) option prefix
      ARG(3) option postfix
      return tag option or ''
   */
   tmp = GetTag(ARG(1))
   IF tmp = '' THEN
      RETURN ''
   RETURN ARG(2)tmp||ARG(3)

/* tagging functions */
GetFormat: PROCEDURE EXPOSE opt.
/* examine data format
   parameter:
      ARG(1) stream
   return:
      'MP3'  MPEG audio
      'OGG'  ogg vorbis
      ''     error
*/
   tmp = CHARIN(ARG(1), 1, 3)
   SELECT
    WHEN tmp = 'ID3' | BITAND(tmp, 'FFF000'x) = 'FFF000'x THEN
      RETURN 'MP3'
    WHEN opt.ogg & tmp = 'Ogg' THEN
      RETURN 'OGG'
    OTHERWISE
      SAY "file "ARG(1)" seems not to have a recognized format. ("C2X(tmp)")"
      RETURN ''
      END

ReadTag: PROCEDURE EXPOSE opt. tag.
/* read file information
   parameter:
      ARG(1) stream
   return:
      format OK
      ''     error
   global:
      tag.*  file information
*/
   /* read header, read tag*/
   CALL STREAM ARG(1), 'c', 'open read'
   fform = GetFormat(ARG(1))
   IF fform = '' THEN DO
      CALL STREAM ARG(1), 'c', 'close'
      RETURN ''
      END
   /* read tag */
   DROP tag.
   SELECT
    WHEN fform = 'MP3' THEN DO /* MP3 */
      /* read ID3 V1.x Tag */
      num = STREAM(ARG(1), 'c', 'QUERY SIZE')
      IF num < 1024 THEN DO
         SAY 'Cannot read tag of file 'ARG(1)'.'
         CALL STREAM ARG(1), 'c', 'close'
         RETURN ''
         END
      IF opt.trunc THEN DO
         tag = CHARIN(ARG(1), num-1023, 1024)
         CALL STREAM ARG(1), 'c', 'close'
         p = VERIFY(REVERSE(tag)' ', '00'x)
         IF p > 1024-128 THEN DO
            SAY 'Cannot find tag of file 'ARG(1)'.'
            RETURN fform
            END
         p = POS('TAG', tag, 1025-128 - p)
         IF p = 0 THEN DO
            SAY 'Cannot find tag of file 'ARG(1)'.'
            RETURN fform
            END
         CALL MMSetFileSize ARG(1), num-1025+p+128
         IF result \= 0 THEN
            SAY 'Failed to truncate 'ARG(1)', error 'result'.'
         tag = SUBSTR(tag, p, 128)
         END
       ELSE DO
         tag = CHARIN(ARG(1), num-127, 128)
         CALL STREAM ARG(1), 'c', 'close'
         IF SUBSTR(tag, 1, 3) \= 'TAG' THEN DO
            SAY 'File 'ARG(1)' has no tag information.'
            RETURN fform
            END
         END
      /* parse tag, CP translation */
      CALL SetTag 'title', SUBSTR(tag, 4, 30)
      CALL SetTag 'artist', SUBSTR(tag, 34, 30)
      CALL SetTag 'album', SUBSTR(tag, 64, 30)
      CALL SetTag 'year', SUBSTR(tag, 94, 4)
      CALL SetTag 'comment', SUBSTR(tag, 98, 30)
      tmp = RIGHT(C2D(SUBSTR(tag, 127, 1)), 2, '0')
      IF tmp \= 0 & SUBSTR(tag, 126, 1) = '00'x THEN
         tag.track = tmp
      tmp = C2D(RIGHT(tag, 1))
      IF tmp \= 255 THEN
         tag.tmp = tmp
      /* done */
      tag.old = 1
      RETURN fform
      END
    WHEN fform = 'OGG' THEN DO
      CALL STREAM ARG(1), 'c', 'close'
      /* launch vorbiscomment */
      tmpfile = SysTempFileName(VALUE('TEMP', ,'OS2ENVIRONMENT')'\mp3t????.tmp')
      '@vorbiscomment 'Escape(ARG(1))' >'tmpfile
      IF RC \= 0 THEN DO
         SAY 'Execution of vorbiscomment returned a bad status ('RC')'
         pause
         RETURN ''
         END
      /* read result */
      IF STREAM(tmpfile, 'c', 'open read') \= 'READY:' THEN DO
         SAY 'Failed to read result from vorbiscomment ('tmpfile')'
         RETURN ''
         END
      r = 0
      DO FOREVER
         tmp = LINEIN(tmpfile)
         IF tmp = '' THEN LEAVE
         r = 1
         PARSE VAR tmp key'='val
         key = STRIP2(key)
         SELECT
          WHEN key = 'title' | key = 'artist' | key = 'album' | key = 'comment' THEN
            CALL VALUE 'tag.'key, val
          WHEN key = 'tracknumber' THEN
            tag.track = RIGHT(val, 2, '0')
          WHEN key = 'date' THEN
            tag.year = val
          OTHERWISE
            END
         END
      CALL STREAM tmpfile, 'c', 'close'
      CALL SysFileDelete(tmpfile)
      /* done */
      IF r = 0 THEN
         SAY 'No comments found in file 'ARG(1)'.'
      RETURN fform
      END
    OTHERWISE NOP
      END
   RETURN ''

WriteTag: PROCEDURE EXPOSE opt. tag.
/* read file information
   parameter:
      ARG(1) stream
      ARG(2) format
   return:
      1      OK
      0      error
   global:
      tag.*  file information
*/
   SELECT
    WHEN ARG(2) = 'MP3' THEN DO /* MP3 */
      num = STREAM(ARG(1), 'c', 'QUERY SIZE')
      IF tag.old = 1 THEN
         num = num -128
      IF STREAM(ARG(1), 'c', 'open write') \= 'READY:' THEN DO
         SAY "file "ARG(1)": failed to write ID3 tag."||'07'x
         RETURN 0
         END
      IF SYMBOL('tag.track') = 'VAR' THEN
         tag2 = LEFT(GetTag('comment'),28)'00'x||D2C(tag.track, 1)
       ELSE
         tag2 = LEFT(GetTag('comment'),30)
      CALL CHAROUT ARG(1), 'TAG'LEFT(GetTag('title'),30)LEFT(GetTag('artist'),30)LEFT(GetTag('album'),30)LEFT(GetTag('year'),4)tag2||D2C(GetTag('style',-1),1), num+1
      CALL STREAM ARG(1), 'c', 'close'
      END
    WHEN ARG(2) = 'OGG' THEN DO
      /* create backup */
      bakfile = ARG(1)'.bak'
      '@COPY 'Escape(ARG(1))' 'Escape(bakfile)
      IF RC \= 0 THEN DO
         SAY 'Failed to create backup copy of file 'ARG(1)'.'
         RETURN 0
         END
      '@vorbiscomment -w 'GetTagOption('artist', '-t "artist=', '" ')GetTagOption('title', '-t "title=', '" ')GetTagOption('album', '-t "album=', '" ')GetTagOption('year', '-t "date=', '" ')GetTagOption('comment', '-t "comment=', '" ')GetTagOption('track', '-t "tracknumber=', '" ')Escape(bakfile)' 'Escape(ARG(1))
      IF RC \= 0 THEN DO
         SAY 'Failed to execute vorbiscomment (File 'ARG(1)').'
         RETURN 0
         END
      CALL SysFileDelete(bakfile)
      END
    OTHERWISE
      RETURN 0
      END
   SAY "File "ARG(1)": tag written."
   RETURN 1

NameEA: PROCEDURE
   IF ARG(2) \= ARG(3) THEN DO
      IF SysPutEA(ARG(1)||ARG(2), '.LONGNAME', 'FDFF'x||D2C(LENGTH(ARG(3)))||'00'x||ARG(3)) \= 0 THEN
         SAY longpath": failed to write .LONGNAME EA."
      END
    ELSE
      CALL SysPutEA ARG(1)||ARG(2), '.LONGNAME', ''
   RETURN

/* replace all components of path ARG(1) by .LONGNAME if exists */
ToLongname: PROCEDURE
   path = Path(ARG(1))
   IF RIGHT(path, 1) \= '\' THEN
      RETURN ARG(1)
   name = FILESPEC('N', ARG(1))
   IF name \= '' THEN
      IF SysGetEA(ARG(1), '.LONGNAME', 'ea') = 0 THEN
         IF ea \= '' THEN
            name = SUBSTR(ea, 5)
   RETURN ToLongname(DELSTR(path, LENGTH(path)))'\'name

CaseAdj: PROCEDURE
   tmp = ARG(1)
   p = 1
   DO FOREVER
      tmp = OVERLAY(MMUpper(SUBSTR(tmp, p, 1)), tmp, p)
      p = VERIFY(tmp, XRANGE('A','Z')XRANGE('a','z')XRANGE('80'x,'FF'x)"0123456789'",, p)
      IF p = 0 | p = LENGTH(tmp) THEN RETURN tmp
      p = p +1
      END

Escape: PROCEDURE
   tmp = ARG(1)
   c = ARG(2)
   IF c = '' THEN c = '%'
   p = POS(c, tmp)
   DO WHILE p \= 0
      tmp = INSERT(c, tmp, p)
      p = POS(c, tmp, p+2)
      END
   RETURN '"'tmp'"'

Path:
   RETURN FILESPEC('D', ARG(1))FILESPEC('P', ARG(1))

Name: PROCEDURE
   n = FILESPEC('N', ARG(1))
   p = LASTPOS('.', n)
   IF p = 0 THEN RETURN n
   RETURN DELSTR(n, p)

Extension: PROCEDURE
   n = FILESPEC('N', ARG(1))
   p = LASTPOS('.', n)
   IF p = 0 THEN RETURN ''
   RETURN SUBSTR(n, p)

/* CP translation ISO to system CP */
Cp2DOS:
   IF \opt.cp THEN RETURN ARG(1)
   RETURN MMTranslateCp(ARG(1), 1004, , '')

/* CP translation system CP to ISO */
Cp2ISO:
   IF \opt.cp THEN RETURN ARG(1)
   RETURN MMTranslateCp(ARG(1), , 1004, '')

/* create filename translation map */
InitTSP: PROCEDURE
   col = SysGetCollate()
   upc = SysMapCase(XRANGE('00'x,'FF'x))
   cvt = ''
   DO i = 0 TO 255
      c = D2C(255)
      IF COMPARE(SUBSTR(col,i+1,1),SUBSTR(upc,i+1,1)) = 0 THEN DO
         p = POS(SUBSTR(col,i+1,1),upc)
         IF p-1 = C2D(SUBSTR(col,i+1,1)) THEN
            p = POS(SUBSTR(col,i+1,1),upc,p+1)
         IF p > 0 THEN
            c = D2C(p-1)
         IF COMPARE(SUBSTR(col,i+1,1), D2C(i)) = 0 & i < 127 & i >= 32 THEN
            c = D2C(i)
         END
       ELSE
         c = SUBSTR(col,i+1,1)
      cvt = cvt||c
      END
   RETURN TRANSLATE(cvt, "'.!;()---", '"*?:<>/\|')

/* translate special characters to similar ones and remove unallowed chars */
TSP: PROCEDURE EXPOSE opt.tsp
   name = STRIP(TRANSLATE(ARG(1), opt.tsp, XRANGE('00'x, 'FF'x)))
   p = 1
   DO FOREVER
      p = POS(D2C(255), name, p)
      IF p = 0 THEN
         RETURN name
      name = DELSTR(name, p, 1)
      END

TSP2:
   RETURN TRANSLATE(TRANSLATE(ARG(1),, XRANGE('80'x, 'FF'x)))

Strip2: PROCEDURE
   tmp = ARG(1)
   opt = TRANSLATE(LEFT(ARG(2), 1, 'B'))
   IF ARG(3, 'e') THEN
      chr = ARG(2)
    ELSE
      chr = '2009'x
   /* strip front */
   IF TRANSLATE(opt, 'L', 'B') = 'L' THEN DO
      p = VERIFY(tmp, chr)
      IF p = 0 THEN
         RETURN ''
       ELSE
         tmp = SUBSTR(tmp, p)
      END
   /* strip back */
   IF TRANSLATE(opt, 'T', 'B') = 'T' THEN DO
      p = VERIFY(REVERSE(tmp), chr)
      RETURN LEFT(tmp, LENGTH(tmp) - p +1)
      END
   RETURN tmp

Error: /* raise error */
   CALL LINEOUT STDERR, ARG(2)
   EXIT ARG(1)

